/*
# $BEGIN_SIGNALACQUISITION_LICENSE$
# 
# This file is part of the SignalAcquisition project, a library for the Arduino
# IDE that samples signals at regular intervals and sends the resulting samples
# in batches over serial port using a simple binary protocol. 
# 
# Author: Jeremy Hill (2023-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SIGNALACQUISITION_LICENSE$
*/

#ifndef    __SignalAcquisition_CPP__
#define    __SignalAcquisition_CPP__

#include "SignalAcquisition.h"

template< typename NUMERICTYPE >
SignalAcquisition< NUMERICTYPE >::
SignalAcquisition( Stream & _stream, unsigned int maxPins ) :
  stream( _stream ),
  samplesPerSecond( 1000.0 ), // default value (zero would disable acquisition)
  samplesPerBlock( 0 ),       // default value (zero disables acquisition)
  sourcePins( "26 27" ),      // default value (on the Pico GP26, GP27, GP28 are ADC pins,
                              // located at physical positions 31,  32,   34 with 33 being the ground,
                              // mapping to breadboard rows    10,   9,    7 with  8 being the ground)
  commOverheadMicros( 0 ),
  idleLoops( 0 ),
  measuredMicrosPerSample( 0 ),
  
  mIdleLoopsSoFar( 0 ),
  mLastSampleTimeInMicros( 0 ), mNextSampleTimeInMicros( 0 ), mNominalMicrosPerSample( 0 ),
  mMaxPins( 0 ), mSourcePinAddresses( 0 ),
  mBufferStart( NULL ), mBufferCursor( NULL ), mBufferFull( NULL ), mBufferSize( 0 )
{
  mMaxPins = maxPins;
  mSourcePinAddresses = mMaxPins ? new int[ mMaxPins ] : NULL;
}

template< typename NUMERICTYPE >
SignalAcquisition< NUMERICTYPE >::
~SignalAcquisition()
{
  delete [] mBufferStart;
  mBufferStart = NULL;
  delete [] mSourcePinAddresses;
  mSourcePinAddresses = NULL;
}

template< typename NUMERICTYPE >
void SignalAcquisition< NUMERICTYPE >::
exposeVariables( Keyhole & keyhole ) 
{ // Call this function inside your `if( keyhole.begin() )` block, and pass that same Keyhole instance
  if( keyhole.variable( "samplesPerSecond", this->samplesPerSecond ) ) configure( keyhole, "samplesPerSecond" );
  if( keyhole.variable( "samplesPerBlock",  this->samplesPerBlock  ) ) configure( keyhole, "samplesPerBlock"  );
  if( keyhole.variable( "sourcePins",       this->sourcePins       ) ) configure( keyhole, "sourcePins"       );

  keyhole.variable( "measuredMicrosPerSample", this->measuredMicrosPerSample, VARIABLE_READ_ONLY );   // for debugging
  keyhole.variable( "idleLoops", this->idleLoops, VARIABLE_READ_ONLY ); // for debugging. Needs to be reliably > 0
  keyhole.variable( "commOverheadMicros", this->commOverheadMicros );   // for debugging. Needs to be reliably < mNominalMicrosPerSample. Set to 0 to disable measurement.
}

template< typename NUMERICTYPE >
void SignalAcquisition< NUMERICTYPE >::
configure( Keyhole & keyhole, const String & reportBadValues )
{
# define CONFIG_FAIL( VAR, INFO ) { if( reportBadValues == VAR || reportBadValues == "*" ) { keyhole.errorStream( "BadValue" ) << INFO; } return; }

  // Tear everything down and start with a blank slate:
  delete [] mBufferStart;
  mBufferStart = mBufferCursor = mBufferFull = NULL;
  mBufferSize = 0;
  for( unsigned int i = 0; i < mMaxPins; i++ )
    mSourcePinAddresses[ i ] = -1; // when Acquire() encounters a -1, it knows that's the end

  // Parse the pin specification (count the pins and store their addresses):
  unsigned int numberOfPins = 0;
  const char * start = this->sourcePins.c_str();
  char * end = NULL;
  while( true )
  {
    while( isspace( *start ) ) start++;
    if( !*start ) break;
    int pin = Keyhole::strToSignedInteger( start, &end );
    if( end && *end && !isspace( *end ) )
      CONFIG_FAIL( "sourcePins", "failed to interpret source pin specification \"" << this->sourcePins << "\" as a space-delimited list of integers" );
    if( pin < 0 )
      CONFIG_FAIL( "sourcePins", "source pin addresses may not be negative" );
    if( numberOfPins == mMaxPins )
      CONFIG_FAIL( "sourcePins", "cannot configure more than " << mMaxPins << " source pins" );
    mSourcePinAddresses[ numberOfPins++ ] = pin;
    preparePin( pin ); // this has been made a virtual method, so that it can be easily overshadowed in a subclass
    start = ( const char * )end;
    end = NULL;
  }
  if( !numberOfPins ) return;
  
  if( this->samplesPerSecond > 0.0 )
  {
    // Configure sample timing 
    mNominalMicrosPerSample = ( int )( 0.5 + 1e6 / this->samplesPerSecond );
    if( abs( 1e6 / this->samplesPerSecond  - mNominalMicrosPerSample ) > 1e-8 )
      CONFIG_FAIL( "samplesPerSecond", "sample period (1e6 / samplesPerSecond) must be a whole number of microseconds" );
    
    // Configure buffer
    if( this->samplesPerBlock > 0 && numberOfPins > 0 )
    {
      // At this point, all parameters are good, so gBufferStart can be non-NULL:
      mBufferStart = new NUMERICTYPE[ this->samplesPerBlock * numberOfPins ];
      mBufferFull = mBufferStart + this->samplesPerBlock * numberOfPins; // NB: pointer arithmetic: sizeof(*mBufferStart) is taken into account
      mBufferCursor = mBufferStart;
      mBufferSize = this->samplesPerBlock * numberOfPins * sizeof( NUMERICTYPE );
    }
    mLastSampleTimeInMicros = micros();
    mNextSampleTimeInMicros = mLastSampleTimeInMicros + mNominalMicrosPerSample;
  }
}

template< typename NUMERICTYPE >
void SignalAcquisition< NUMERICTYPE >::
reset( unsigned int currentTimeInMicros )
{
  mLastSampleTimeInMicros = currentTimeInMicros;
  mNextSampleTimeInMicros = mLastSampleTimeInMicros + mNominalMicrosPerSample;
  mBufferCursor = mBufferStart;
}

template< typename NUMERICTYPE >
void SignalAcquisition< NUMERICTYPE >::
acquire( unsigned int currentTimeInMicros )
{
  // If acquisition hasn't been properly configured, abort:
  if( !mBufferStart ) return;
  
  // If it isn't time for a sample yet, abort:
  if( ( int )( currentTimeInMicros - mNextSampleTimeInMicros ) < 0 ) { mIdleLoopsSoFar++; return; }
  // don't rearrange the arithmetic, it's wraparound-safe the way it is

  // Read each pin once - that's one sample:
  readSample(); // this has been made a virtual method, so that it can be easily overshadowed in a subclass
    
  // If the sample-block is now complete, send it:
  if( mBufferCursor == mBufferFull )
  {
    unsigned int t0 = this->commOverheadMicros ? micros() : 0;
    // First send a sample-block header, which begins with a 16-bit
    // integer with value 1. This will be received as either \x00\x01
    // or \x01\x00 : either of these will (a) signal to the receiver that
    // a sample-block is coming, and (b) reveal the sender's endianity.
    const uint16_t endianMarker = 1;
    this->stream.write( ( uint8_t * )&endianMarker, sizeof( endianMarker ) );
    this->stream.println(); // a line-ending is also required to complete the header
    // Now send the actual sample-block data:
    this->stream.write( ( uint8_t * )mBufferStart, mBufferSize );
    this->stream.println(); // this line-ending is optional, but makes things easier in the Arduino IDE Serial Monitor
    this->stream.flush();   // this is needed to ensure tight timing
    mBufferCursor = mBufferStart; // empties the buffer
    if( this->commOverheadMicros ) this->commOverheadMicros = micros() - t0;
    // set commOverheadMicros to 0 if you want to prevent it being measured (saves the overhead of two calls to micros())
  }
  
  // Record the time taken, empirically, between the last sample and this one
  // (this variable is exposed via Keyhole, so you can set samplesPerSecond too high
  // and see where this measure levels off).
  this->measuredMicrosPerSample = currentTimeInMicros - mLastSampleTimeInMicros;
  mLastSampleTimeInMicros = currentTimeInMicros;
  
  this->idleLoops = mIdleLoopsSoFar;
  mIdleLoopsSoFar = 0; // you should call DoAcquisition() on every loop(), but some calls
  // will be aborted by the logic above. The idleLoops counter, exposed as "idleLoops",
  // counts the aborted loops. If the count ever dips to 0, that's a sign that your
  // sketch may not be able to keep up with the desired sampling rate.
  
  // Schedule the next sample:
  mNextSampleTimeInMicros += mNominalMicrosPerSample;
  // If we're behind, don't get cumulatively further behind (otherwise the time may
  // eventually wrap around and we'll stop realizing we're behind). This means
  // we get graceful behavior even if samplesPerSecond is set too high:
  if( ( int )( currentTimeInMicros - mNextSampleTimeInMicros ) > 0 ) mNextSampleTimeInMicros = currentTimeInMicros;
  // don't rearrange the arithmetic, it's wraparound-safe the way it is
}

// With template classes, you get "undefined reference" errors from the linker unless you either
// (1) in the .h file, include the actual definitions (.cpp content) or (2) in the .cpp file,
// declare `template class SignalAcquisition<X>;` for each allowable X.  Solution (2) makes for
// faster compilation but may make the compiled sketch a few bytes bigger on some boards (e.g.,
// +64 bytes on ItsyBitsy M4, or +72 on Raspberry Pi Pico) while making no difference on some
// other boards (Teensy 3.1, Arduino UNO). See https://stackoverflow.com/a/8752879
template class SignalAcquisition<float>;
template class SignalAcquisition<double>;
template class SignalAcquisition<int16_t>;
template class SignalAcquisition<uint16_t>;
template class SignalAcquisition<int32_t>;
template class SignalAcquisition<uint32_t>;

#endif // __SignalAcquisition_CPP__
